from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from typing import TypeVar, Generic, Union, List, Sequence, Optional

from metricflow.column_assoc import ColumnAssociation, SingleColumnCorrelationKey
from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dag.id_generation import IdGeneratorRegistry
from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNodeVisitor,
    FilterElementsNode,
    WriteToResultDataframeNode,
    OrderByLimitNode,
    ComputeMetricsNode,
    AggregateMeasuresNode,
    JoinAggregatedMeasuresByGroupByColumnsNode,
    JoinToBaseOutputNode,
    ReadSqlSourceNode,
    BaseOutput,
    ComputedMetricsOutput,
    WhereConstraintNode,
    CombineMetricsNode,
    SourceDataSetT,
    ConstrainTimeRangeNode,
    WriteToResultTableNode,
    JoinOverTimeRangeNode,
    MetricTimeDimensionTransformNode,
)
from metricflow.dataset.dataset import DataSet
from metricflow.instances import (
    AggregationState,
    InstanceSet,
    MetricInstance,
    MetricModelReference,
    TimeDimensionInstance,
)
from metricflow.model.objects.metric import CumulativeMetricWindow, MetricType
from metricflow.model.semantic_model import SemanticModel
from metricflow.object_utils import assert_values_exhausted
from metricflow.plan_conversion.instance_converters import (
    RemoveMeasures,
    AddMetrics,
    CreateSelectColumnsForInstances,
    CreateSelectColumnsWithMeasuresAggregated,
    create_select_columns_for_instance_sets,
    AddLinkToLinkableElements,
    FilterElements,
    ChangeAssociatedColumns,
    ChangeMeasureAggregationState,
    FilterLinkableInstancesWithLeadingLink,
)
from metricflow.plan_conversion.select_column_gen import (
    SelectColumnSet,
)
from metricflow.plan_conversion.spec_transforms import (
    CreateOnConditionForCombiningMetrics,
    CreateSelectCoalescedColumnsForLinkableSpecs,
    SelectOnlyLinkableSpecs,
)
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlEngineAttributes, SupportedSqlEngine
from metricflow.specs import (
    ColumnAssociationResolver,
    MetricSpec,
    TimeDimensionSpec,
    MeasureSpec,
)
from metricflow.sql.optimizer.optimization_levels import (
    SqlQueryOptimizationLevel,
    SqlQueryOptimizerConfiguration,
)
from metricflow.sql.sql_exprs import (
    SqlExpressionNode,
    SqlComparisonExpression,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlComparison,
    SqlIsNullExpression,
    SqlLogicalExpression,
    SqlLogicalOperator,
    SqlStringExpression,
    SqlCastToTimestampExpression,
    SqlTimeDeltaExpression,
    SqlRatioComputationExpression,
    SqlDateTruncExpression,
    SqlStringLiteralExpression,
    SqlBetweenExpression,
)
from metricflow.sql.sql_plan import (
    SqlQueryPlan,
    SqlSelectStatementNode,
    SqlJoinDescription,
    SqlSelectColumn,
    SqlOrderByDescription,
    SqlJoinType,
    SqlQueryPlanNode,
    SqlTableFromClauseNode,
)
from metricflow.time.time_constants import ISO8601_PYTHON_FORMAT
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


# The type of data set that present at a source node.
SqlDataSetT = TypeVar("SqlDataSetT", bound=SqlDataSet)


@dataclass(frozen=True)
class ColumnEqualityDescription:
    """Helper class to enumerate columns that should be equal between sources in a join."""

    left_column_alias: str
    right_column_alias: str


def make_equijoin_description(
    right_data_set: SqlDataSet,
    left_source_alias: str,
    right_source_alias: str,
    column_equality_descriptions: Sequence[ColumnEqualityDescription],
) -> SqlJoinDescription:
    """Make a join description where the condition is an equals between two columns in the left and right sources.

    A source is defined as either a table or a subquery.
    """
    assert len(column_equality_descriptions) > 0

    and_conditions = []
    for column_equality_description in column_equality_descriptions:
        and_conditions.append(
            SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=left_source_alias,
                        column_name=column_equality_description.left_column_alias,
                    )
                ),
                comparison=SqlComparison.EQUALS,
                right_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=right_source_alias,
                        column_name=column_equality_description.right_column_alias,
                    )
                ),
            )
        )
    on_condition: SqlExpressionNode
    if len(and_conditions) == 1:
        on_condition = and_conditions[0]
    else:
        on_condition = SqlLogicalExpression(operator=SqlLogicalOperator.AND, args=tuple(and_conditions))

    return SqlJoinDescription(
        right_source=right_data_set.sql_select_node,
        right_source_alias=right_source_alias,
        on_condition=on_condition,
        join_type=SqlJoinType.LEFT_OUTER,
    )


def _make_aggregate_measures_join_condition(
    column_aliases: Sequence[str],
    left_source_alias: str,
    right_source_alias: str,
) -> SqlLogicalExpression:
    """Make the ON condition particular to the aggregate measures join scenario

    In most cases equality will suffice, but in the event of NULL values on both sides we include
    those in the join. This is questionable, so maybe we shouldn't do it.
    """

    join_conditions = []
    for column_alias in column_aliases:
        left_column = SqlColumnReferenceExpression(
            SqlColumnReference(table_alias=left_source_alias, column_name=column_alias)
        )
        right_column = SqlColumnReferenceExpression(
            SqlColumnReference(table_alias=right_source_alias, column_name=column_alias)
        )
        equality_condition = SqlComparisonExpression(
            left_expr=left_column, right_expr=right_column, comparison=SqlComparison.EQUALS
        )

        left_is_null = SqlIsNullExpression(arg=left_column)
        right_is_null = SqlIsNullExpression(arg=right_column)
        both_null_condition = SqlLogicalExpression(operator=SqlLogicalOperator.AND, args=(left_is_null, right_is_null))

        join_conditions.append(
            SqlLogicalExpression(operator=SqlLogicalOperator.OR, args=(equality_condition, both_null_condition))
        )

    return SqlLogicalExpression(
        operator=SqlLogicalOperator.AND,
        args=tuple(join_conditions),
    )


def _make_cumulative_metric_join_condition(
    from_data_set_alias: str,
    from_data_set_identifier_col: str,
    right_data_set_alias: str,
    right_data_set_identifier_col: str,
    window: Optional[CumulativeMetricWindow],
) -> SqlExpressionNode:
    """Creates SqlLogicalExpression representing a cumulative metric self-join condition"""
    if window is None:  # accumulate over all time
        return SqlComparisonExpression(
            left_expr=SqlColumnReferenceExpression(
                SqlColumnReference(
                    table_alias=from_data_set_alias,
                    column_name=from_data_set_identifier_col,
                )
            ),
            comparison=SqlComparison.LESS_THAN_OR_EQUALS,
            right_expr=SqlColumnReferenceExpression(
                SqlColumnReference(
                    table_alias=right_data_set_alias,
                    column_name=right_data_set_identifier_col,
                )
            ),
        )

    # ds <= <date> and  ds > <date> - <window>
    return SqlLogicalExpression(
        operator=SqlLogicalOperator.AND,
        args=(
            SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=from_data_set_identifier_col,
                    )
                ),
                comparison=SqlComparison.LESS_THAN_OR_EQUALS,
                right_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=right_data_set_alias,
                        column_name=right_data_set_identifier_col,
                    )
                ),
            ),
            SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=from_data_set_identifier_col,
                    )
                ),
                comparison=SqlComparison.GREATER_THAN,
                right_expr=SqlTimeDeltaExpression(
                    SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=right_data_set_alias,
                            column_name=right_data_set_identifier_col,
                        )
                    ),
                    count=window.count,
                    granularity=window.granularity,
                    grain_to_date=None,
                ),
            ),
        ),
    )


def _make_grain_to_date_cumulative_metric_join_condition(
    from_data_set_alias: str,
    from_data_set_identifier_col: str,
    right_data_set_alias: str,
    right_data_set_identifier_col: str,
    grain_to_date: TimeGranularity,
) -> SqlLogicalExpression:
    """Creates SqlLogicalExpression representing a grain_to_date cumulative metric self-join condition"""
    return SqlLogicalExpression(
        operator=SqlLogicalOperator.AND,
        args=(
            SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=from_data_set_identifier_col,
                    )
                ),
                comparison=SqlComparison.LESS_THAN_OR_EQUALS,
                right_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=right_data_set_alias,
                        column_name=right_data_set_identifier_col,
                    )
                ),
            ),
            SqlComparisonExpression(
                left_expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=from_data_set_identifier_col,
                    )
                ),
                comparison=SqlComparison.GREATER_THAN_OR_EQUALS,
                right_expr=SqlTimeDeltaExpression(
                    SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=right_data_set_alias,
                            column_name=right_data_set_identifier_col,
                        )
                    ),
                    count=0,
                    granularity=grain_to_date,
                    grain_to_date=grain_to_date,
                ),
            ),
        ),
    )


def _make_time_range_comparison_expr(
    table_alias: str, column_alias: str, time_range_constraint: TimeRangeConstraint
) -> SqlExpressionNode:
    """Build an expression like "ds BETWEEN CAST('2020-01-01' AS TIMESTAMP) AND CAST('2020-01-02' AS TIMESTAMP)"""
    # TODO: Update when adding < day granularity support.
    return SqlBetweenExpression(
        column_arg=SqlColumnReferenceExpression(
            SqlColumnReference(
                table_alias=table_alias,
                column_name=column_alias,
            )
        ),
        start_expr=SqlCastToTimestampExpression(
            arg=SqlStringLiteralExpression(
                literal_value=time_range_constraint.start_time.strftime(ISO8601_PYTHON_FORMAT),
            )
        ),
        end_expr=SqlCastToTimestampExpression(
            arg=SqlStringLiteralExpression(
                literal_value=time_range_constraint.end_time.strftime(ISO8601_PYTHON_FORMAT),
            )
        ),
    )


def _make_time_spine_data_set(
    metric_time_dimension_instance: TimeDimensionInstance,
    metric_time_dimension_column_name: str,
    time_spine_source: TimeSpineSource,
    time_spine_table_alias: str,
    time_range_constraint: Optional[TimeRangeConstraint] = None,
) -> SqlDataSet:
    """Make a time spine data set, which contains all date values like '2020-01-01', '2020-01-02'...

    This is useful in computing cumulative metrics. This will need to be updated to support granularities finer than a
    day.
    """
    time_spine_instance = (
        TimeDimensionInstance(
            defined_from=metric_time_dimension_instance.defined_from,
            associated_columns=(
                ColumnAssociation(
                    column_name=metric_time_dimension_column_name,
                    column_correlation_key=SingleColumnCorrelationKey(),
                ),
            ),
            spec=metric_time_dimension_instance.spec,
        ),
    )
    time_spine_instance_set = InstanceSet(
        time_dimension_instances=time_spine_instance,
    )
    description = "Date Spine"

    # If the requested granularity is the same as the granularity of the spine, do a direct select.
    if metric_time_dimension_instance.spec.time_granularity == time_spine_source.time_column_granularity:
        return SqlDataSet(
            instance_set=time_spine_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=(
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            SqlColumnReference(
                                table_alias=time_spine_table_alias,
                                column_name=time_spine_source.time_column_name,
                            ),
                        ),
                        column_alias=metric_time_dimension_column_name,
                    ),
                ),
                from_source=SqlTableFromClauseNode(sql_table=time_spine_source.spine_table),
                from_source_alias=time_spine_table_alias,
                joins_descs=(),
                group_bys=(),
                where=_make_time_range_comparison_expr(
                    table_alias=time_spine_table_alias,
                    column_alias=time_spine_source.time_column_name,
                    time_range_constraint=time_range_constraint,
                )
                if time_range_constraint
                else None,
                order_bys=(),
            ),
        )
    # If the granularity is different, apply a DATE_TRUNC() and aggregate.
    else:
        select_columns = (
            SqlSelectColumn(
                expr=SqlDateTruncExpression(
                    time_granularity=metric_time_dimension_instance.spec.time_granularity,
                    arg=SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=time_spine_table_alias,
                            column_name=time_spine_source.time_column_name,
                        ),
                    ),
                ),
                column_alias=metric_time_dimension_column_name,
            ),
        )
        return SqlDataSet(
            instance_set=time_spine_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=select_columns,
                from_source=SqlTableFromClauseNode(sql_table=time_spine_source.spine_table),
                from_source_alias=time_spine_table_alias,
                joins_descs=(),
                group_bys=select_columns,
                where=_make_time_range_comparison_expr(
                    table_alias=time_spine_table_alias,
                    column_alias=time_spine_source.time_column_name,
                    time_range_constraint=time_range_constraint,
                )
                if time_range_constraint
                else None,
                order_bys=(),
            ),
        )


class DataflowToSqlQueryPlanConverter(Generic[SqlDataSetT], DataflowPlanNodeVisitor[SqlDataSetT, SqlDataSet]):
    """Generates an SQL query plan from a node in the a metric dataflow plan."""

    def __init__(
        self,
        column_association_resolver: ColumnAssociationResolver,
        semantic_model: SemanticModel,
        time_spine_source: TimeSpineSource,
    ) -> None:
        """Constructor.

        Args:
            column_association_resolver: controls how columns for instances are generated and used between nested
            queries.
            semantic_model: Self-explanatory.
            time_spine_source: Allows getting dates for use in cumulative joins
        """
        self._column_association_resolver = column_association_resolver
        self._metric_semantics = semantic_model.metric_semantics
        self._data_source_semantics = semantic_model.data_source_semantics
        self._time_spine_source = time_spine_source

    @property
    def column_association_resolver(self) -> ColumnAssociationResolver:  # noqa: D
        return self._column_association_resolver

    def _next_unique_table_alias(self) -> str:
        """Return the next unique table alias to use in generating queries."""
        return IdGeneratorRegistry.for_class(self.__class__).create_id(prefix="subq")

    def visit_source_node(self, node: ReadSqlSourceNode[SqlDataSetT]) -> SqlDataSet:
        """Generate the SQL to read from the source."""
        return SqlDataSet(
            sql_select_node=node.data_set.sql_select_node,
            instance_set=node.data_set.instance_set,
        )

    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode[SqlDataSetT]) -> SqlDataSet:
        """Generate time range join SQL"""
        table_alias_to_instance_set: OrderedDict[str, InstanceSet] = OrderedDict()

        input_data_set = node.parent_node.accept(self)
        input_data_set_alias = self._next_unique_table_alias()

        sql_join_descs: List[SqlJoinDescription] = []

        metric_time_dimension_spec: Optional[TimeDimensionSpec] = None
        metric_time_dimension_instance: Optional[TimeDimensionInstance] = None
        for instance in input_data_set.instance_set.time_dimension_instances:
            if (
                instance.spec.element_name == node.metric_time_dimension_reference.element_name
                and len(instance.spec.identifier_links) == 0
            ):
                metric_time_dimension_instance = instance
                metric_time_dimension_spec = instance.spec
                break

        # If the metric time dimension isn't present in the parent node it's because it wasn't requested
        # and therefore we don't need the time range join because we can just let the metric sum over all time
        if metric_time_dimension_spec is None:
            return input_data_set

        input_data_set_identifier_column_association = input_data_set.column_association_for_time_dimension(
            metric_time_dimension_spec
        )
        input_data_set_identifier_col = input_data_set_identifier_column_association.column_name

        time_spine_data_set_alias = self._next_unique_table_alias()

        metric_time_dimension_column_name = self.column_association_resolver.resolve_time_dimension_spec(
            metric_time_dimension_spec
        ).column_name

        # assemble dataset with metric_time_dimension to join
        assert metric_time_dimension_instance
        time_spine_data_set = _make_time_spine_data_set(
            metric_time_dimension_instance=metric_time_dimension_instance,
            metric_time_dimension_column_name=metric_time_dimension_column_name,
            time_spine_source=self._time_spine_source,
            time_spine_table_alias=self._next_unique_table_alias(),
            time_range_constraint=node.time_range_constraint,
        )
        table_alias_to_instance_set[time_spine_data_set_alias] = time_spine_data_set.instance_set

        # Figure out which columns in the "right" data set correspond to the time dimension that we want to join on.
        time_spine_data_set_column_associations = time_spine_data_set.column_association_for_time_dimension(
            metric_time_dimension_spec
        )
        time_spine_data_set_time_dimension_col = time_spine_data_set_column_associations.column_name

        # Build an expression like "a.ds <= b.ds AND a.ds >= b.ds - <window>
        # If no window is present we join across all time -> "a.ds <= b.ds"
        constrain_metric_time_column_condition_both: Optional[SqlExpressionNode] = None
        if node.window is not None:
            constrain_metric_time_column_condition_both = _make_cumulative_metric_join_condition(
                input_data_set_alias,
                input_data_set_identifier_col,
                time_spine_data_set_alias,
                time_spine_data_set_time_dimension_col,
                node.window,
            )
        elif node.grain_to_date is not None:
            constrain_metric_time_column_condition_both = _make_grain_to_date_cumulative_metric_join_condition(
                input_data_set_alias,
                input_data_set_identifier_col,
                time_spine_data_set_alias,
                time_spine_data_set_time_dimension_col,
                node.grain_to_date,
            )
        elif node.window is None:  # `window is None` for clarity (could be else since we have window and !window)
            constrain_metric_time_column_condition_both = _make_cumulative_metric_join_condition(
                input_data_set_alias,
                input_data_set_identifier_col,
                time_spine_data_set_alias,
                time_spine_data_set_time_dimension_col,
                None,
            )

        sql_join_descs.append(
            SqlJoinDescription(
                right_source=input_data_set.sql_select_node,
                right_source_alias=input_data_set_alias,
                on_condition=constrain_metric_time_column_condition_both,
                join_type=SqlJoinType.INNER,
            )
        )

        modified_input_instance_set = InstanceSet(
            measure_instances=input_data_set.instance_set.measure_instances,
            dimension_instances=input_data_set.instance_set.dimension_instances,
            identifier_instances=input_data_set.instance_set.identifier_instances,
            metric_instances=input_data_set.instance_set.metric_instances,
            # we omit the metric time dimension from the right side of the self-join because we need to use
            # the metric time dimension from the right side
            time_dimension_instances=tuple(
                [
                    time_dimension_instance
                    for time_dimension_instance in input_data_set.instance_set.time_dimension_instances
                    if time_dimension_instance.spec != metric_time_dimension_spec
                ]
            ),
        )
        table_alias_to_instance_set[input_data_set_alias] = modified_input_instance_set

        # The output instances are the same as the input instances.
        output_instance_set = ChangeAssociatedColumns(self._column_association_resolver).transform(
            input_data_set.instance_set
        )

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=create_select_columns_for_instance_sets(
                    self._column_association_resolver, table_alias_to_instance_set
                ),
                from_source=time_spine_data_set.sql_select_node,
                from_source_alias=time_spine_data_set_alias,
                joins_descs=tuple(sql_join_descs),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_join_to_base_output_node(self, node: JoinToBaseOutputNode[SqlDataSetT]) -> SqlDataSet:
        """Generates the query that realizes the behavior of the JoinToStandardOutputNode."""

        # Keep a mapping between the table aliases that would be used in the query and the MDO instances in that source.
        # e.g. when building "FROM from_table a JOIN right_table b", the value for key "a" would be the instances in
        # "from_table"
        table_alias_to_instance_set: OrderedDict[str, InstanceSet] = OrderedDict()

        # Convert the dataflow from the left node to a DataSet and add context for it to table_alias_to_instance_set
        # A DataSet is a bundle of the SQL query (in object form) and the MDO instances that the SQL query contains.
        from_data_set = node.left_node.accept(self)
        from_data_set_alias = self._next_unique_table_alias()
        table_alias_to_instance_set[from_data_set_alias] = from_data_set.instance_set

        # Build the join descriptions for the SqlQueryPlan - different from node.join_descriptions which are the join
        # descriptions from the dataflow plan.
        sql_join_descs: List[SqlJoinDescription] = []

        # The dataflow plan describes how the data sets coming from the parent nodes should be joined together. Use
        # those descriptions to convert them to join descriptions for the SQL query plan.
        for join_description in node.join_targets:
            join_on_identifier = join_description.join_on_identifier

            # Figure out which columns in the "from" data set correspond to the identifier that we want to join on.
            # The column associations tell us which columns correspond to which instances in the data set.
            from_data_set_identifier_column_associations = from_data_set.column_associations_for_identifier(
                join_on_identifier
            )
            from_data_set_identifier_cols = [c.column_name for c in from_data_set_identifier_column_associations]

            right_node_to_join: BaseOutput = join_description.join_node
            right_data_set: SqlDataSet = right_node_to_join.accept(self)
            right_data_set_alias = self._next_unique_table_alias()

            # Figure out which columns in the "right" data set correspond to the identifier that we want to join on.
            right_data_set_column_associations = right_data_set.column_associations_for_identifier(join_on_identifier)
            right_data_set_identifier_cols = [c.column_name for c in right_data_set_column_associations]

            assert len(from_data_set_identifier_cols) == len(
                right_data_set_identifier_cols
            ), f"Cannot construct join - the number of columns on the left ({from_data_set_identifier_cols}) side of the join does not match the right ({right_data_set_identifier_cols})"

            # We have the columns that we need to "join on" in the query, so add it to the list of join descriptions to
            # use later.
            column_equality_descriptions = []
            for idx in range(len(from_data_set_identifier_cols)):
                column_equality_descriptions.append(
                    ColumnEqualityDescription(
                        left_column_alias=from_data_set_identifier_cols[idx],
                        right_column_alias=right_data_set_identifier_cols[idx],
                    )
                )
            # Add the partition columns as well.
            for dimension_join_description in join_description.join_on_partition_dimensions:
                column_equality_descriptions.append(
                    ColumnEqualityDescription(
                        left_column_alias=from_data_set.column_association_for_dimension(
                            dimension_join_description.start_node_dimension_spec
                        ).column_name,
                        right_column_alias=right_data_set.column_association_for_dimension(
                            dimension_join_description.node_to_join_dimension_spec
                        ).column_name,
                    )
                )

            for time_dimension_join_description in join_description.join_on_partition_time_dimensions:
                column_equality_descriptions.append(
                    ColumnEqualityDescription(
                        left_column_alias=from_data_set.column_association_for_time_dimension(
                            time_dimension_join_description.start_node_time_dimension_spec
                        ).column_name,
                        right_column_alias=right_data_set.column_association_for_time_dimension(
                            time_dimension_join_description.node_to_join_time_dimension_spec
                        ).column_name,
                    )
                )

            sql_join_descs.append(
                make_equijoin_description(
                    right_data_set=right_data_set,
                    left_source_alias=from_data_set_alias,
                    right_source_alias=right_data_set_alias,
                    column_equality_descriptions=column_equality_descriptions,
                )
            )

            # Remove the linkable instances with the join_on_identifier as the leading link as the next step adds the
            # link. This is to avoid cases where there is a primary identifier and a dimension in the data set, and we
            # create an instance in the next step that has the same identifier link.
            # e.g. a data set has the dimension "listing__country_latest" and "listing" is a primary identifier in the
            # data set. The next step would create an instance like "listing__listing__country_latest" without this
            # filter.

            # logger.error(f"before filter is:\n{pformat_big_objects(right_data_set.instance_set.spec_set)}")
            right_data_set_instance_set_filtered = FilterLinkableInstancesWithLeadingLink(
                identifier_link=join_on_identifier,
            ).transform(right_data_set.instance_set)
            # logger.error(f"after filter is:\n{pformat_big_objects(right_data_set_instance_set_filtered.spec_set)}")

            # After the right data set is joined to the "from" data set, we need to change the links for some of the
            # instances that represent the right data set. For example, if the "from" data set contains the "bookings"
            # measure instance and the right dataset contains the "country" dimension instance, then after the join,
            # the output data set should have the "country" dimension instance with the "user_id" identifier link
            # (if "user_id" equality was the join condition). "country" -> "user_id__country"
            right_data_set_instance_set_after_join = right_data_set_instance_set_filtered.transform(
                AddLinkToLinkableElements(join_on_identifier=join_on_identifier)
            )
            table_alias_to_instance_set[right_data_set_alias] = right_data_set_instance_set_after_join

        from_data_set_output_instance_set = from_data_set.instance_set.transform(
            FilterElements(include_specs=from_data_set.instance_set.spec_set.all_specs)
        )

        # Change the aggregation state for the measures to be partially aggregated if it was previously aggregated
        # since we removed the identifiers and added the dimensions. The dimensions could have the same value for
        # multiple rows, so we'll need to re-aggregate.
        from_data_set_output_instance_set = from_data_set_output_instance_set.transform(
            ChangeMeasureAggregationState(
                {
                    AggregationState.NON_AGGREGATED: AggregationState.NON_AGGREGATED,
                    AggregationState.COMPLETE: AggregationState.PARTIAL,
                    AggregationState.PARTIAL: AggregationState.PARTIAL,
                }
            )
        )

        table_alias_to_instance_set[from_data_set_alias] = from_data_set_output_instance_set

        # Construct the data set that contains the updated instances and the SQL nodes that should go in the various
        # clauses.
        return SqlDataSet(
            instance_set=InstanceSet.merge(list(table_alias_to_instance_set.values())),
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=create_select_columns_for_instance_sets(
                    self._column_association_resolver, table_alias_to_instance_set
                ),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=tuple(sql_join_descs),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_join_aggregated_measures_by_groupby_columns_node(  # noqa: D
        self, node: JoinAggregatedMeasuresByGroupByColumnsNode
    ) -> SqlDataSet:
        """Generates the query that realizes the behavior of the JoinAggregatedMeasuresByGroupByColumnsNode.

        This node is a straight inner join against all of the columns used for grouping in the input
        aggregation steps. Every column should be used, and at this point all inputs are fully aggregated,
        meaning we can make assumptions about things like NULL handling and there being one row per value
        set in each data source.

        In addition, this is used in cases where we expect a final metric to be computed using these
        measures as input. Therefore, we make the assumption that any mismatch should be discarded, as
        the behavior of the metric will be undefined in that case. This is why the INNER JOIN type is
        appropriate - if a dimension value set exists in one aggregated set but not the other, there is
        no sensible metric value for that dimension set.
        """
        assert len(node.parent_nodes) > 1, "This cannot happen, the node initializer would have failed"

        table_alias_to_instance_set: OrderedDict[str, InstanceSet] = OrderedDict()

        from_data_set: SqlDataSet = node.parent_nodes[0].accept(self)
        from_data_set_alias = self._next_unique_table_alias()
        table_alias_to_instance_set[from_data_set_alias] = from_data_set.instance_set
        join_aliases = [column.column_name for column in from_data_set.groupable_column_associations]

        sql_join_descs: List[SqlJoinDescription] = []
        for aggregated_node in node.parent_nodes[1:]:
            right_data_set: SqlDataSet = aggregated_node.accept(self)
            right_data_set_alias = self._next_unique_table_alias()
            right_column_names = {column.column_name for column in right_data_set.groupable_column_associations}
            if right_column_names != set(join_aliases):
                # TODO test multi-hop dimensions and address any issues. For now, let's raise an exception.
                raise ValueError(
                    f"We only support joins where all dimensions have the same name, but we got {right_column_names} "
                    f"and {join_aliases}, which differ by {right_column_names.difference(set(join_aliases))}!"
                )
            sql_join_descs.append(
                SqlJoinDescription(
                    right_source=right_data_set.sql_select_node,
                    right_source_alias=right_data_set_alias,
                    on_condition=_make_aggregate_measures_join_condition(
                        column_aliases=join_aliases,
                        left_source_alias=from_data_set_alias,
                        right_source_alias=right_data_set_alias,
                    ),
                    join_type=SqlJoinType.INNER,
                )
            )
            # All groupby columns are shared by all inputs, so we only want the measure/metric columns
            # from the data sources on the right side of the join
            table_alias_to_instance_set[right_data_set_alias] = InstanceSet(
                measure_instances=right_data_set.instance_set.measure_instances,
                metric_instances=right_data_set.instance_set.metric_instances,
            )

        return SqlDataSet(
            instance_set=InstanceSet.merge(list(table_alias_to_instance_set.values())),
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=create_select_columns_for_instance_sets(
                    self._column_association_resolver, table_alias_to_instance_set
                ),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=tuple(sql_join_descs),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode) -> SqlDataSet:
        """Generates the query that realizes the behavior of AggregateMeasuresNode."""
        # Get the data from the parent, and change measure instances to the aggregated state.
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        aggregated_instance_set = from_data_set.instance_set.transform(
            ChangeMeasureAggregationState(
                {
                    AggregationState.NON_AGGREGATED: AggregationState.COMPLETE,
                    AggregationState.COMPLETE: AggregationState.COMPLETE,
                    AggregationState.PARTIAL: AggregationState.COMPLETE,
                }
            )
        )
        # Also, the columns should always follow the resolver format.
        aggregated_instance_set = aggregated_instance_set.transform(
            ChangeAssociatedColumns(self._column_association_resolver)
        )

        from_data_set_alias = self._next_unique_table_alias()

        # The call below does a few things, but there are a few function calls that convert the instance set into
        # SQL query plan objects.
        select_column_set: SelectColumnSet = aggregated_instance_set.transform(
            CreateSelectColumnsWithMeasuresAggregated(
                from_data_set_alias,
                self._column_association_resolver,
                self._data_source_semantics,
            )
        )
        return SqlDataSet(
            instance_set=aggregated_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This will generate expressions with the appropriate aggregation functions e.g. SUM()
                select_columns=select_column_set.as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                # This will generate expressions to group by the columns that don't correspond to a measure instance.
                group_bys=select_column_set.without_measure_columns().as_tuple(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> SqlDataSet:
        """Generates the query that realizes the behavior of ComputeMetricsNode."""
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        from_data_set_alias = self._next_unique_table_alias()

        # TODO: Check that all measures for the metrics are in the input instance set
        # The desired output instance set has no measures, so create a copy with those removed.
        output_instance_set: InstanceSet = from_data_set.instance_set.transform(RemoveMeasures())

        # Also, the output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        non_metric_select_column_set: SelectColumnSet = output_instance_set.transform(
            CreateSelectColumnsForInstances(
                table_alias=from_data_set_alias,
                column_resolver=self._column_association_resolver,
            )
        )

        # Add select columns that would compute the metrics to the select columns.
        metric_select_columns = []
        metric_instances = []
        for metric_spec in node.metric_specs:
            metric = self._metric_semantics.get_metric(metric_spec)

            metric_expr: Optional[SqlExpressionNode] = None
            if metric.type is MetricType.RATIO:
                numerator = metric.type_params.numerator_measure_reference
                denominator = metric.type_params.denominator_measure_reference
                assert (
                    numerator is not None and denominator is not None
                ), "Missing numerator or denominator for ratio metric, this should have been caught in validation!"
                numerator_column_name = self._column_association_resolver.resolve_measure_spec(
                    MeasureSpec(element_name=numerator.element_name)
                ).column_name
                denominator_column_name = self._column_association_resolver.resolve_measure_spec(
                    MeasureSpec(element_name=denominator.element_name)
                ).column_name

                metric_expr = SqlRatioComputationExpression(
                    numerator=SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=from_data_set_alias,
                            column_name=numerator_column_name,
                        )
                    ),
                    denominator=SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=from_data_set_alias,
                            column_name=denominator_column_name,
                        )
                    ),
                )
            elif metric.type is MetricType.MEASURE_PROXY:
                if len(metric.measure_references) > 0:
                    assert (
                        len(metric.measure_references) == 1
                    ), "Measure proxy metrics should always source from exactly 1 measure."
                    expr = self._column_association_resolver.resolve_measure_spec(
                        MeasureSpec(element_name=metric.measure_references[0].element_name)
                    ).column_name
                else:
                    expr = metric.name
                # Use a column reference to improve query optimization.
                metric_expr = SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=expr,
                    )
                )
            elif metric.type is MetricType.CUMULATIVE:
                assert (
                    len(metric.measure_references) == 1
                ), "Cumulative metrics should always source from exactly 1 measure."
                expr = self._column_association_resolver.resolve_measure_spec(
                    MeasureSpec(element_name=metric.measure_references[0].element_name)
                ).column_name
                metric_expr = SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=expr,
                    )
                )
            elif metric.type == MetricType.EXPR:
                assert metric.type_params.expr
                metric_expr = SqlStringExpression(sql_expr=metric.type_params.expr)
            else:
                assert_values_exhausted(metric.type)

            assert metric_expr

            output_column_association = self._column_association_resolver.resolve_metric_spec(metric_spec)
            metric_select_columns.append(
                SqlSelectColumn(
                    expr=metric_expr,
                    column_alias=output_column_association.column_name,
                )
            )
            metric_instances.append(
                MetricInstance(
                    associated_columns=(output_column_association,),
                    defined_from=[MetricModelReference(metric_name=metric_spec.element_name)],
                    spec=metric_spec,
                )
            )
        output_instance_set = output_instance_set.transform(AddMetrics(metric_instances))

        combined_select_column_set = non_metric_select_column_set.merge(
            SelectColumnSet(metric_columns=metric_select_columns)
        )

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=combined_select_column_set.as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> SqlDataSet:  # noqa: D
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        output_instance_set = from_data_set.instance_set
        from_data_set_alias = self._next_unique_table_alias()

        # Also, the output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        order_by_descriptions = []
        for order_by_spec in node.order_by_specs:
            column_associations = order_by_spec.item.column_associations(self._column_association_resolver)
            for column_association in column_associations:
                order_by_descriptions.append(
                    SqlOrderByDescription(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias=from_data_set_alias, column_name=column_association.column_name
                            )
                        ),
                        desc=order_by_spec.descending,
                    )
                )

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=None,
                order_bys=tuple(order_by_descriptions),
                limit=node.limit,
            ),
        )

    def visit_write_to_result_dataframe_node(self, node: WriteToResultDataframeNode) -> SqlDataSet:
        """This is an operation that can't be represented as an SQL query.

        Instead, it should be handled in the execution plan as an operation that runs an SQL query and saves it to
        a dataframe.
        """
        raise RuntimeError("This node type is not supported.")

    def visit_write_to_result_table_node(self, node: WriteToResultTableNode[SourceDataSetT]) -> SqlDataSet:
        """Similar to visit_write_to_result_dataframe_node()."""
        raise RuntimeError("This node type is not supported.")

    def visit_pass_elements_filter_node(self, node: FilterElementsNode) -> SqlDataSet:
        """Generates the query that realizes the behavior of FilterElementsNode."""
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        output_instance_set = from_data_set.instance_set.transform(FilterElements(node.include_specs))
        from_data_set_alias = self._next_unique_table_alias()

        # Also, the output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_where_constraint_node(self, node: WhereConstraintNode) -> SqlDataSet:
        """Adds where clause to SQL statement from parent node"""
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        output_instance_set = from_data_set.instance_set
        from_data_set_alias = self._next_unique_table_alias()

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=SqlStringExpression(
                    sql_expr=node.where.where_condition,
                    used_columns=tuple(node.where.linkable_names),
                    execution_parameters=node.where.execution_parameters,
                ),
                order_bys=(),
            ),
        )

    def _make_select_columns_for_metrics(
        self, table_alias_to_metric_specs: OrderedDict[str, Sequence[MetricSpec]]
    ) -> List[SqlSelectColumn]:
        """Creates select columns that get the given metric using the given table alias.

        e.g.

        with table_alias_to_metric_specs = {"a": MetricSpec(element_name="bookings")}

        ->

        a.bookings AS bookings
        """
        select_columns = []
        for table_alias, metric_specs in table_alias_to_metric_specs.items():
            for metric_spec in metric_specs:
                metric_column_name = self._column_association_resolver.resolve_metric_spec(metric_spec).column_name
                select_columns.append(
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias=table_alias,
                                column_name=metric_column_name,
                            )
                        ),
                        column_alias=metric_column_name,
                    )
                )
        return select_columns

    def visit_combine_metrics_node(self, node: CombineMetricsNode[SqlDataSetT]) -> SqlDataSet:  # noqa: D
        # Sanity check that all parents have the same linkable specs.
        parent_data_sets = [x.accept(self) for x in node.parent_nodes]
        assert (
            len(parent_data_sets) > 1
        ), "Shouldn't have a CombineMetricsNode in the dataflow plan if there's only 1 parent."
        linkable_specs = parent_data_sets[0].instance_set.spec_set.linkable_specs
        assert all(
            [set(x.instance_set.spec_set.linkable_specs) == set(linkable_specs) for x in parent_data_sets]
        ), "All parent nodes should have the same set of linkable instances since all values are coalesced."
        linkable_spec_set = parent_data_sets[0].instance_set.spec_set.transform(SelectOnlyLinkableSpecs())

        # Create a FULL OUTER join where the join key is all previous dimension values from all sources coalesced.
        #
        # e.g.
        #
        # FROM (
        #   ...
        # ) subq_9
        # FULL OUTER JOIN (
        #   ...
        # ) subq_10
        # ON
        #   subq_9.is_instant = subq_10.is_instant
        #   AND subq_9.ds = subq_10.ds
        # FULL OUTER JOIN (
        #   ...
        # ) subq_11
        # ON
        #   COALESCE(subq_9.is_instant, subq_10.is_instant) = subq_11.is_instant
        #   AND COALESCE(subq_9.ds, subq_10.ds) = subq_11.ds

        joins_descriptions = []
        parent_source_table_aliases = []
        table_alias_to_metric_specs: OrderedDict[str, Sequence[MetricSpec]] = OrderedDict()
        for parent_data_set in parent_data_sets:
            table_alias = self._next_unique_table_alias()
            parent_source_table_aliases.append(table_alias)
            table_alias_to_metric_specs[table_alias] = parent_data_set.instance_set.spec_set.metric_specs

        # Create the components of the FULL OUTER JOIN that combines metrics. Order doesn't
        # matter so we use the first element in the FROM clause and create join descriptions from
        # the rest.
        from_alias = parent_source_table_aliases[0]
        from_source = parent_data_sets[0].sql_select_node
        for i in range(1, len(parent_source_table_aliases)):
            joins_descriptions.append(
                SqlJoinDescription(
                    right_source=parent_data_sets[i].sql_select_node,
                    right_source_alias=parent_source_table_aliases[i],
                    on_condition=CreateOnConditionForCombiningMetrics(
                        column_association_resolver=self._column_association_resolver,
                        table_aliases_in_coalesce=parent_source_table_aliases[:i],
                        table_alias_on_right_equality=parent_source_table_aliases[i],
                    ).transform(linkable_spec_set),
                    join_type=SqlJoinType.FULL_OUTER,
                )
            )

        # We can merge all parent instances since the common linkable instances will be de-duped.
        output_instance_set = InstanceSet.merge([x.instance_set for x in parent_data_sets])
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        metric_select_column_set = SelectColumnSet(
            metric_columns=self._make_select_columns_for_metrics(table_alias_to_metric_specs)
        )
        linkable_select_column_set = linkable_spec_set.transform(
            CreateSelectCoalescedColumnsForLinkableSpecs(
                column_association_resolver=self._column_association_resolver,
                table_aliases=parent_source_table_aliases,
            )
        )
        combined_select_column_set = linkable_select_column_set.merge(metric_select_column_set)

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=combined_select_column_set.as_tuple(),
                from_source=from_source,
                from_source_alias=from_alias,
                joins_descs=tuple(joins_descriptions),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode[SourceDataSetT]) -> SqlDataSet:  # noqa: D
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        from_data_set_alias = self._next_unique_table_alias()

        time_dimension_instances_for_metric_time = [
            time_dimension_instance
            for time_dimension_instance in from_data_set.instance_set.time_dimension_instances
            if time_dimension_instance.spec.element_name == DataSet.metric_time_dimension_name()
        ]

        # Use the smallest time granularity to build the comparison since that's what was used in the data source
        # definition and it wouldn't have a DATE_TRUNC() in the expression. We want to build
        #
        # ds >= '2020-01-01' AND ds <= '2020-02-01'
        #
        # not DATE_TRUNC('month', ds) >= '2020-01-01' AND DATE_TRUNC('month', ds <= '2020-02-01')
        def sort_function_for_time_granularity(instance: TimeDimensionInstance) -> int:
            return instance.spec.time_granularity.to_int()

        time_dimension_instances_for_metric_time.sort(key=sort_function_for_time_granularity)
        assert (
            len(time_dimension_instances_for_metric_time) > 0
        ), "No metric time dimensions found in the input data set for this node"

        time_dimension_instance_for_metric_time = time_dimension_instances_for_metric_time[0]

        # Build an expression like "ds >= CAST('2020-01-01' AS TIMESTAMP) AND ds <= CAST('2020-01-02' AS TIMESTAMP)"
        constrain_metric_time_column_condition = _make_time_range_comparison_expr(
            table_alias=from_data_set_alias,
            column_alias=time_dimension_instance_for_metric_time.associated_column.column_name,
            time_range_constraint=node.time_range_constraint,
        )

        output_instance_set = from_data_set.instance_set
        # Output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=constrain_metric_time_column_condition,
                order_bys=(),
            ),
        )

    def convert_to_sql_query_plan(
        self,
        sql_engine_attributes: SqlEngineAttributes,
        sql_query_plan_id: str,
        dataflow_plan_node: Union[BaseOutput, ComputedMetricsOutput],
        optimization_level: SqlQueryOptimizationLevel = SqlQueryOptimizationLevel.O4,
    ) -> SqlQueryPlan:
        """Create an SQL query plan that represents the computation up to the given dataflow plan node."""

        sql_select_node: SqlQueryPlanNode = dataflow_plan_node.accept(self).sql_select_node

        # TODO: Make this a more generally accessible attribute instead of checking against the
        # BigQuery-ness of the engine
        use_column_alias_in_group_by = sql_engine_attributes.sql_engine_type is SupportedSqlEngine.BIGQUERY

        for optimizer in SqlQueryOptimizerConfiguration.optimizers_for_level(
            optimization_level, use_column_alias_in_group_by=use_column_alias_in_group_by
        ):
            logger.info(f"Applying optimizer: {optimizer.__class__.__name__}")
            sql_select_node = optimizer.optimize(sql_select_node)

        return SqlQueryPlan(plan_id=sql_query_plan_id, render_node=sql_select_node)

    def visit_metric_time_dimension_transform_node(
        self, node: MetricTimeDimensionTransformNode[SqlDataSetT]
    ) -> SqlDataSet:
        """Implement the behavior of the MetricTimeDimensionTransformNode.

        This node will create an output data set that is similar to the input data set, but the measure instances it
        contains is a subset of the input data set. Only measure instances that have an aggregation time dimension
        matching the one defined in the node will be passed. In addition, an additional time dimension instance for
        "metric time" will be included. See DataSet.metric_time_dimension_reference().
        """
        input_data_set: SqlDataSet = node.parent_node.accept(self)

        # Find which measures have an aggregation time dimension that is the same as the one specified in the node.
        # Only these measures will be in the output data set.
        output_measure_instances = []
        for measure_instance in input_data_set.instance_set.measure_instances:
            measure = self._data_source_semantics.get_measure(measure_instance.spec.as_reference)
            if measure.checked_agg_time_dimension == node.aggregation_time_dimension_reference:
                output_measure_instances.append(measure_instance)

        if len(output_measure_instances) == 0:
            raise RuntimeError(
                f"No measure instances in the input source match the aggregation time dimension "
                f"{node.aggregation_time_dimension_reference}. Check if the dataflow plan was constructed correctly."
            )

        # Find time dimension instances that refer to the same dimension as the one specified in the node.
        matching_time_dimension_instances = []
        for time_dimension_instance in input_data_set.instance_set.time_dimension_instances:
            # The specification for the time dimension to use for aggregation is the local one.
            if (
                len(time_dimension_instance.spec.identifier_links) == 0
                and time_dimension_instance.spec.reference == node.aggregation_time_dimension_reference
            ):
                matching_time_dimension_instances.append(time_dimension_instance)

        output_time_dimension_instances: List[TimeDimensionInstance] = []
        output_time_dimension_instances.extend(input_data_set.instance_set.time_dimension_instances)
        output_column_to_input_column: OrderedDict[str, str] = OrderedDict()

        # For those matching time dimension instances, create the analog metric time dimension instances for the output.
        for matching_time_dimension_instance in matching_time_dimension_instances:
            metric_time_dimension_spec = DataSet.metric_time_dimension_spec(
                matching_time_dimension_instance.spec.time_granularity
            )
            metric_time_dimension_column_association = self._column_association_resolver.resolve_time_dimension_spec(
                metric_time_dimension_spec
            )
            output_time_dimension_instances.append(
                TimeDimensionInstance(
                    defined_from=matching_time_dimension_instance.defined_from,
                    associated_columns=(
                        self._column_association_resolver.resolve_time_dimension_spec(metric_time_dimension_spec),
                    ),
                    spec=metric_time_dimension_spec,
                )
            )
            output_column_to_input_column[
                metric_time_dimension_column_association.column_name
            ] = matching_time_dimension_instance.associated_column.column_name
        output_instance_set = InstanceSet(
            measure_instances=tuple(output_measure_instances),
            dimension_instances=input_data_set.instance_set.dimension_instances,
            time_dimension_instances=tuple(output_time_dimension_instances),
            identifier_instances=input_data_set.instance_set.identifier_instances,
            metric_instances=input_data_set.instance_set.metric_instances,
        )
        output_instance_set = ChangeAssociatedColumns(self._column_association_resolver).transform(output_instance_set)

        from_data_set_alias = self._next_unique_table_alias()

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=CreateSelectColumnsForInstances(
                    column_resolver=self._column_association_resolver,
                    table_alias=from_data_set_alias,
                    output_to_input_column_mapping=output_column_to_input_column,
                )
                .transform(output_instance_set)
                .as_tuple(),
                from_source=input_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )



from typing import Dict, List
from metricflow.instances import DataSourceElementReference

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.validations.validator_helpers import (
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    ModelValidationRule,
    DimensionInvariants,
    ValidationIssueType,
    ValidationError,
    validate_safely,
)
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.specs import DimensionReference
from metricflow.time.time_granularity import TimeGranularity


class DimensionConsistencyRule(ModelValidationRule):
    """Checks for consistent dimension properties in the data sources in a model.

    * Dimensions with the same name should be of the same type.
    * Dimensions with the same name should be either all partitions or not.
    """

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring dimension consistency")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        dimension_to_invariant: Dict[DimensionReference, DimensionInvariants] = {}
        time_dims_to_granularity: Dict[DimensionReference, TimeGranularity] = {}
        issues: List[ValidationIssueType] = []

        for data_source in model.data_sources:
            issues += DimensionConsistencyRule._validate_data_source(
                data_source=data_source, dimension_to_invariant=dimension_to_invariant, update_invariant_dict=True
            )

            for dimension in data_source.dimensions:
                issues += DimensionConsistencyRule._validate_dimension(
                    dimension=dimension,
                    time_dims_to_granularity=time_dims_to_granularity,
                    data_source=data_source,
                )
        return issues

    @staticmethod
    @validate_safely(
        whats_being_done="checking that time dimensions of the same name that are not primary "
        "have the same time granularity specifications"
    )
    def _validate_dimension(
        dimension: Dimension,
        time_dims_to_granularity: Dict[DimensionReference, TimeGranularity],
        data_source: DataSource,
    ) -> List[ValidationIssueType]:
        """Checks that time dimensions of the same name that aren't primary have the same time granularity specifications

        Args:
            dimension: the dimension to check
            time_dims_to_granularity: a dict from the dimension to the time granularity it should have
            data_source: the associated data source. Used for generated issue messages
        Throws: MdoValidationError if there is an inconsistent dimension in the data source.
        """
        issues: List[ValidationIssueType] = []
        context = DataSourceElementContext(
            file_context=FileContext.from_metadata(metadata=data_source.metadata),
            data_source_element=DataSourceElementReference(
                data_source_name=data_source.name, element_name=dimension.name
            ),
            element_type=DataSourceElementType.DIMENSION,
        )

        if dimension.type == DimensionType.TIME:
            if dimension.reference not in time_dims_to_granularity and dimension.type_params:
                time_dims_to_granularity[dimension.reference] = dimension.type_params.time_granularity

                # The primary time dimension can be of different time granularities, so don't check for it.
                if (
                    dimension.type_params is not None
                    and not dimension.type_params.is_primary
                    and dimension.type_params.time_granularity != time_dims_to_granularity[dimension.reference]
                ):
                    expected_granularity = time_dims_to_granularity[dimension.reference]
                    issues.append(
                        ValidationError(
                            context=context,
                            message=f"Time granularity must be the same for time dimensions with the same name. "
                            f"Problematic dimension: {dimension.name} in data source with name: "
                            f"`{data_source.name}`. Expected granularity is {expected_granularity.name}.",
                        )
                    )

        return issues

    @staticmethod
    @validate_safely(
        whats_being_done="checking that the data source has dimensions consistent with the given invariants"
    )
    def _validate_data_source(
        data_source: DataSource,
        dimension_to_invariant: Dict[DimensionReference, DimensionInvariants],
        update_invariant_dict: bool,
    ) -> List[ValidationIssueType]:
        """Checks that the given data source has dimensions consistent with the given invariants.

        Args:
            data_source: the data source to check
            dimension_to_invariant: a dict from the dimension name to the properties it should have
            update_invariant_dict: whether to insert an entry into the dict if the given dimension name doesn't exist.
        Throws: MdoValidationError if there is an inconsistent dimension in the data source.
        """
        issues: List[ValidationIssueType] = []

        for dimension in data_source.dimensions:
            dimension_invariant = dimension_to_invariant.get(dimension.reference)

            if dimension_invariant is None:
                if update_invariant_dict:
                    dimension_invariant = DimensionInvariants(dimension.type, dimension.is_partition or False)
                    dimension_to_invariant[dimension.reference] = dimension_invariant
                    continue
                # TODO: Can't check for unknown dimensions easily as the name follows <id>__<name> format.
                # e.g. user__created_at
                continue

            # is_partition might not be specified in the configs, so default to False.
            is_partition = dimension.is_partition or False

            context = DataSourceElementContext(
                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                data_source_element=DataSourceElementReference(
                    data_source_name=data_source.name, element_name=dimension.name
                ),
                element_type=DataSourceElementType.DIMENSION,
            )

            if dimension_invariant.type != dimension.type:
                issues.append(
                    ValidationError(
                        context=context,
                        message=f"In data source `{data_source.name}`, type conflict for dimension `{dimension.name}` "
                        f"- already in model as type `{dimension_invariant.type}` but got `{dimension.type}`",
                    )
                )
            if dimension_invariant.is_partition != is_partition:
                issues.append(
                    ValidationError(
                        context=context,
                        message=f"In data source `{data_source.name}, conflicting is_partition attribute for dimension "
                        f"`{dimension.reference}` - already in model"
                        f" with is_partition as `{dimension_invariant.is_partition}` but got "
                        f"`{is_partition}``",
                    )
                )

        return issues



import pytest
from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.optimizer.sub_query_reducer import SqlSubQueryReducer
from metricflow.sql.sql_exprs import (
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlComparisonExpression,
    SqlComparison,
)
from metricflow.sql.sql_plan import (
    SqlSelectStatementNode,
    SqlSelectColumn,
    SqlTableFromClauseNode,
    SqlOrderByDescription,
    SqlJoinDescription,
    SqlJoinType,
)
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.sql.compare_sql_plan import assert_default_rendered_sql_equal


@pytest.fixture
def base_select_statement() -> SqlSelectStatementNode:
    """SELECT statement used to build test cases.

    SELECT
      src2.col0 AS col0
      src2.col1 AS col1
      -- src2
      SELECT
        src1.col0 AS col0
        , src1.col1 AS col1
      FROM (
        -- src1
        SELECT
          src0.col0 AS col0
          , src0.col1 AS col1
        FROM demo.from_source_table src0
        LIMIT 2
      ) src1
      LIMIT 1
    ) src2
    ORDER BY src2.col0
    """
    return SqlSelectStatementNode(
        description="src3",
        select_columns=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(col_ref=SqlColumnReference(table_alias="src2", column_name="col0")),
                column_alias="col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(col_ref=SqlColumnReference(table_alias="src2", column_name="col1")),
                column_alias="col1",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="src2",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="src1", column_name="col0")
                    ),
                    column_alias="col0",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="src1", column_name="col1")
                    ),
                    column_alias="col1",
                ),
            ),
            from_source=SqlSelectStatementNode(
                description="src1",
                select_columns=(
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias="src0",
                                column_name="col0",
                            )
                        ),
                        column_alias="col0",
                    ),
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias="src0",
                                column_name="col1",
                            )
                        ),
                        column_alias="col1",
                    ),
                ),
                from_source=SqlTableFromClauseNode(
                    sql_table=SqlTable(schema_name="demo", table_name="from_source_table")
                ),
                from_source_alias="src0",
                joins_descs=(),
                group_bys=(),
                order_bys=(),
                limit=2,
            ),
            from_source_alias="src1",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
            limit=1,
        ),
        from_source_alias="src2",
        joins_descs=(),
        group_bys=(),
        where=None,
        order_bys=(
            SqlOrderByDescription(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="src2",
                        column_name="col0",
                    )
                ),
                desc=False,
            ),
        ),
        limit=None,
    )


def test_reduce_sub_query(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where an outer query should be reduced into its inner query with merged LIMIT expressions"""
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=base_select_statement,
        plan_id="before_reducing",
    )

    sub_query_reducer = SqlSubQueryReducer()

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=sub_query_reducer.optimize(base_select_statement),
        plan_id="after_reducing",
    )


@pytest.fixture
def rewrite_order_by_statement() -> SqlSelectStatementNode:
    """Helps check order-by rewrites when a query is reduced.

    --
    SELECT
      src2.col0
      src2.col1
    FROM (
      SELECT
        src0.col0 AS col0
        src1.col1 AS col1
      FROM demo.src0 src0
      JOIN demo.src1 src1
      ON
        src0.join_col = src1.join_col
    ) src2
    ORDER BY
    src2.col1
    """
    return SqlSelectStatementNode(
        description="src3",
        select_columns=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(col_ref=SqlColumnReference(table_alias="src2", column_name="col0")),
                column_alias="col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(col_ref=SqlColumnReference(table_alias="src2", column_name="col1")),
                column_alias="col1",
            ),
        ),
        from_source=(
            SqlSelectStatementNode(
                description="src2",
                select_columns=(
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(table_alias="src0", column_name="col0")
                        ),
                        column_alias="col0",
                    ),
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(table_alias="src1", column_name="col1")
                        ),
                        column_alias="col1",
                    ),
                ),
                from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="src0")),
                from_source_alias="src0",
                joins_descs=(
                    SqlJoinDescription(
                        right_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="src1")),
                        right_source_alias="src1",
                        on_condition=SqlComparisonExpression(
                            left_expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="src0", column_name="join_col")
                            ),
                            comparison=SqlComparison.EQUALS,
                            right_expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(table_alias="src1", column_name="join_col")
                            ),
                        ),
                        join_type=SqlJoinType.INNER,
                    ),
                ),
                where=None,
                group_bys=(),
                order_bys=(),
            )
        ),
        from_source_alias="src2",
        joins_descs=(),
        group_bys=(),
        where=None,
        order_bys=(
            SqlOrderByDescription(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias="src2",
                        column_name="col1",
                    )
                ),
                desc=False,
            ),
        ),
        limit=None,
    )


def test_rewrite_order_by_with_a_join_in_parent(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    rewrite_order_by_statement: SqlSelectStatementNode,
) -> None:
    """Tests rewriting an order by when the parent has a join."""

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=rewrite_order_by_statement,
        plan_id="before_reducing",
    )

    sub_query_reducer = SqlSubQueryReducer()

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=sub_query_reducer.optimize(rewrite_order_by_statement),
        plan_id="after_reducing",
    )



import logging
import textwrap
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Any, Tuple

import jinja2
import pandas as pd

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlEngineAttributes, SqlClient
from metricflow.sql.sql_bind_parameters import SqlBindParameters

logger = logging.getLogger(__name__)


class SqlClientException(Exception):
    """Raised when an interaction with the SQL engine has an error."""

    pass


class BaseSqlClientImplementation(ABC, SqlClient):
    """Abstract implementation that other SQL clients are based on."""

    INDENT = "    "

    def generate_health_check_tests(self, schema_name: str) -> List[Tuple[str, Any]]:  # type: ignore
        """List of base health checks we want to perform."""
        table_name = "health_report"
        return [
            ("SELECT 1", lambda: self.execute("SELECT 1")),
            (f"Create schema '{schema_name}'", lambda: self.create_schema(schema_name)),
            (
                f"Create table '{schema_name}.{table_name}' with a SELECT",
                lambda: self.create_table_as_select(
                    SqlTable(schema_name=schema_name, table_name="health_report"), "SELECT 'test' AS test_col"
                ),
            ),
            (
                f"Drop table '{schema_name}.{table_name}'",
                lambda: self.drop_table(SqlTable(schema_name=schema_name, table_name="health_report")),
            ),
        ]

    def health_checks(self, schema_name: str) -> Dict[str, Dict[str, str]]:
        """Perform health checks"""
        checks_to_run = self.generate_health_check_tests(schema_name)
        results: Dict[str, Dict[str, str]] = {}

        for step, check in checks_to_run:
            status = "SUCCESS"
            err_string = ""
            try:
                resp = check()
                logger.info(f"Health Check Item {step}: succeeded" + f" with response {str(resp)}" if resp else None)
            except Exception as e:
                status = "FAIL"
                err_string = str(e)
                logger.error(f"Health Check Item {step}: failed with error {err_string}")

            results[f"{type(self).__name__.lower()} - {step}"] = {
                "status": status,
                "error_message": err_string,
            }

        return results

    def query(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> pd.DataFrame:
        """Query statement; result expected to be data which will be returned as a DataFrame

        Args:
            stmt: The SQL query statement to run. This should produce output via a SELECT
            sql_bind_parameters: The parameter replacement mapping for filling in
                concrete values for SQL query parameters.
        """
        start = time.time()
        logger.info(
            f"Running query:"
            f"\n\n{textwrap.indent(stmt, prefix=BaseSqlClientImplementation.INDENT)}\n"
            + (f"\nwith parameters: {dict(sql_bind_parameters.param_dict)}" if sql_bind_parameters.param_dict else "")
        )
        df = self._engine_specific_query_implementation(stmt, sql_bind_parameters)
        if not isinstance(df, pd.DataFrame):
            raise RuntimeError(f"Expected query to return a DataFrame, got {type(df)}")
        stop = time.time()
        logger.info(f"Finished running the query in {stop - start:.2f}s with {df.shape[0]} row(s) returned")
        return df

    def execute(  # noqa: D
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        start = time.time()
        logger.info(
            f"Running query:"
            f"\n\n{textwrap.indent(stmt, prefix=BaseSqlClientImplementation.INDENT)}\n"
            + (f"\nwith parameters: {dict(sql_bind_parameters.param_dict)}" if sql_bind_parameters.param_dict else "")
        )
        self._engine_specific_execute_implementation(stmt, sql_bind_parameters)
        stop = time.time()
        logger.info(f"Finished running the query in {stop - start:.2f}s")
        return None

    def dry_run(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        """Dry run statement; checks that the 'stmt' is queryable. Returns None. Raises an exception if the 'stmt' isn't queryable.

        Args:
            stmt: The SQL query statement to dry run.
            sql_bind_parameters: The parameter replacement mapping for filling in
                concrete values for SQL query parameters.
        """
        start = time.time()
        logger.info(
            f"Running dry_run of:"
            f"\n\n{textwrap.indent(stmt, prefix=BaseSqlClientImplementation.INDENT)}\n"
            + (f"\nwith parameters: {dict(sql_bind_parameters.param_dict)}" if sql_bind_parameters.param_dict else "")
        )
        results = self._engine_specific_dry_run_implementation(stmt, sql_bind_parameters)
        stop = time.time()
        logger.info(f"Finished running the dry_run in {stop - start:.2f}s")
        return results

    @property
    @abstractmethod
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Struct of SQL engine attributes.

        This is used by MetricFlow for things like SQL rendering and safe use of multi-threading.
        """
        raise NotImplementedError

    @abstractmethod
    def _engine_specific_query_implementation(self, stmt: str, bind_params: SqlBindParameters) -> pd.DataFrame:
        """Sub-classes should implement this to query the engine."""
        pass

    @abstractmethod
    def _engine_specific_execute_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:
        """Sub-classes should implement this to execute a statement that doesn't return results."""
        pass

    @abstractmethod
    def _engine_specific_dry_run_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:
        """Sub-classes should implement this to check a query will run successfully without actually running the query"""
        pass

    @abstractmethod
    def create_table_from_dataframe(  # noqa: D
        self,
        sql_table: SqlTable,
        df: pd.DataFrame,
        chunk_size: Optional[int] = None,
    ) -> None:
        pass

    @abstractmethod
    def list_tables(self, schema_name: str) -> List[str]:  # noqa: D
        pass

    def table_exists(self, sql_table: SqlTable) -> bool:  # noqa: D
        return sql_table.table_name in self.list_tables(sql_table.schema_name)

    def create_table_as_select(  # noqa: D
        self,
        sql_table: SqlTable,
        select_query: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        self.execute(
            jinja2.Template(
                textwrap.dedent(
                    """\
                    CREATE TABLE {{ sql_table }} AS
                      {{ select_query | indent(2) }}
                    """
                ),
                undefined=jinja2.StrictUndefined,
            ).render(
                sql_table=sql_table.sql,
                select_query=select_query,
            ),
            sql_bind_parameters=sql_bind_parameters,
        )

    def create_schema(self, schema_name: str) -> None:  # noqa: D
        self.execute(f"CREATE SCHEMA IF NOT EXISTS {schema_name}")

    def drop_schema(self, schema_name: str, cascade: bool = True) -> None:  # noqa: D
        self.execute(f"DROP SCHEMA IF EXISTS {schema_name}{' CASCADE' if cascade else ''}")

    def drop_table(self, sql_table: SqlTable) -> None:  # noqa: D
        self.execute(f"DROP TABLE IF EXISTS {sql_table.sql}")

    def close(self) -> None:  # noqa: D
        pass



from metricflow.api.metricflow_client import MetricFlowClient
from metricflow.dataflow.sql_table import SqlTable
from metricflow.engine.models import Dimension, Materialization, Metric
from metricflow.model.validations.validator_helpers import ModelValidationResults
from metricflow.object_utils import random_id


def test_query(mf_client: MetricFlowClient) -> None:  # noqa: D
    result = mf_client.query(
        ["bookings"],
        ["ds"],
        limit=2,
        start_time="2019-01-01",
        end_time="2024-01-01",
    )
    assert result.query_spec
    assert result.dataflow_plan
    assert result.sql
    assert result.result_df is not None
    assert len(result.result_df) == 2
    assert result.result_table is None

    output_table = SqlTable(schema_name=mf_client.system_schema, table_name=f"test_table_{random_id()}")
    result = mf_client.query(
        ["bookings"], ["ds"], limit=2, start_time="2019-01-01", end_time="2024-01-01", as_table=output_table.sql
    )
    assert result.query_spec
    assert result.dataflow_plan
    assert result.sql
    assert result.result_df is None
    assert result.result_table == output_table


def test_explain(mf_client: MetricFlowClient) -> None:  # noqa: D
    result = mf_client.explain(
        ["bookings"],
        ["ds"],
        limit=2,
        start_time="2019-01-01",
        end_time="2024-01-01",
    )
    assert result.query_spec
    assert result.dataflow_plan
    assert result.execution_plan
    assert result.output_table is None

    output_table = SqlTable(schema_name=mf_client.system_schema, table_name=f"test_table_{random_id()}")
    result = mf_client.explain(
        ["bookings"], ["ds"], limit=2, start_time="2019-01-01", end_time="2024-01-01", as_table=output_table.sql
    )
    assert result.query_spec
    assert result.dataflow_plan
    assert result.execution_plan
    assert result.output_table == output_table


def test_list_metrics(mf_client: MetricFlowClient) -> None:  # noqa: D
    metrics = mf_client.list_metrics()
    assert metrics

    metric_name, metric_obj = next(iter(metrics.items()))

    assert metric_name == metric_obj.name
    assert isinstance(metric_obj, Metric)

    assert metric_obj.dimensions
    assert isinstance(metric_obj.dimensions[0], Dimension)


def test_list_dimensions(mf_client: MetricFlowClient) -> None:  # noqa: D
    dimensions = mf_client.list_dimensions(["bookings"])

    assert dimensions
    assert isinstance(dimensions[0], Dimension)

    dimensions = mf_client.list_dimensions(["bookings", "revenue"])
    assert len(dimensions) == 1
    assert dimensions[0].name == "ds"


def test_list_materializations(mf_client: MetricFlowClient) -> None:  # noqa: D
    materializations = mf_client.list_materializations()
    assert materializations
    assert isinstance(materializations[0], Materialization)


def test_get_dimension_values(mf_client: MetricFlowClient) -> None:  # noqa: D
    dim_vals = mf_client.get_dimension_values("bookings", "ds", start_time="2020-01-01", end_time="2024-01-01")
    assert dim_vals


def test_materializations(mf_client: MetricFlowClient) -> None:  # noqa: D
    mat_name = "test_materialization"
    output_table = mf_client.materialize(materialization_name=mat_name, start_time="2020-01-01", end_time="2024-01-01")
    assert isinstance(output_table, SqlTable)
    assert output_table.sql == f"{mf_client.system_schema}.{mat_name}"

    dropped = mf_client.drop_materialization(mat_name)
    assert dropped


def test_validate_configs(mf_client: MetricFlowClient) -> None:  # noqa: D
    issues = mf_client.validate_configs()
    assert isinstance(issues, ModelValidationResults)



import copy
import datetime
import logging

import dateutil.parser
from _pytest.fixtures import FixtureRequest
from typing import Callable, Tuple, List

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.materialization import Materialization
from metricflow.model.objects.metric import Metric
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.protocols.sql_client import SqlClient

logger = logging.getLogger(__name__)


def as_datetime(date_string: str) -> datetime.datetime:
    """Helper to convert a string like '2020-01-01' into a datetime object."""
    return dateutil.parser.parse(date_string)


def should_skip_multi_threaded(
    request: FixtureRequest,
    sql_client: SqlClient,
) -> bool:
    """Returns whether to skip this test because multi-threading is not supported by the DB"""
    if not sql_client.sql_engine_attributes.multi_threading_supported:
        logger.warning(
            f"Multi-threading is not supported with {sql_client.__class__.__name__}, so should skip "
            f"{request.node.fspath}::{request.node.name}`"
        )
        return True

    return False


def find_data_source_with(model: UserConfiguredModel, function: Callable[[DataSource], bool]) -> Tuple[DataSource, int]:
    """Returns a data source from the model which matches the criteria defined by the passed in function'

    This is useful because the order of data sources in the list is non determinant, thus it's impossible to
    hard code which data source you want by index. Using data source names isn't great for consistency because
    data sources change and might no longer have the necessary parts to be useful for a given test. This
    allows us to guarantee that a data source will be returned which meets the requirements of what a test needs,
    unless none of the data sources will work.
    """
    for index, data_source in enumerate(model.data_sources):
        if function(data_source):
            return data_source, index

    raise Exception("Unable to find a data_source matching function criteria")


def find_metric_with(model: UserConfiguredModel, function: Callable[[Metric], bool]) -> Tuple[Metric, int]:
    """Returns a metric from the model which matches the criteria defined by the passed in function'

    This is useful because the order of metrics in the list is non-determinant, thus it's impossible to
    hard code which metric you want by index. Using metric names isn't great for consistency because
    metrics change and might no longer have the necessary parts to be useful for a given test. This
    allows us to guarantee that a metric will be returned which meets the requirements of what a test needs,
    unless none of the metrics will work.
    """
    for index, metric in enumerate(model.metrics):
        if function(metric):
            return metric, index

    raise Exception("Unable to find a metric matching function criteria")


def model_with_materialization(
    initial_model: UserConfiguredModel, materializations: List[Materialization]
) -> UserConfiguredModel:
    """Returns a copy of the model but with the materialization added"""
    model_copy = copy.deepcopy(initial_model)
    model_copy.materializations.extend(materializations)
    return model_copy



from typing import Sequence, List

from metricflow.dataflow.dataflow_plan import ReadSqlSourceNode, MetricTimeDimensionTransformNode, BaseOutput
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.model.semantic_model import SemanticModel


class SourceNodeBuilder:
    """Helps build source nodes to use in the dataflow plan builder.

    The current use case is for creating a set of input nodes from a data set to support multiple aggregation time
    dimensions. Each data set is converted into k DataFlowPlan nodes, one per distinct aggregation time dimension used
    in the definition of a measure.
    """

    def __init__(self, semantic_model: SemanticModel) -> None:  # noqa: D
        self._semantic_model = semantic_model

    def create_from_data_sets(self, data_sets: Sequence[DataSourceDataSet]) -> Sequence[BaseOutput[DataSourceDataSet]]:
        """Creates source nodes from DataSourceDataSets."""
        source_nodes: List[BaseOutput[DataSourceDataSet]] = []
        for data_set in data_sets:
            read_node = ReadSqlSourceNode[DataSourceDataSet](data_set)
            aggregation_time_dimensions = self._semantic_model.data_source_semantics.get_aggregation_time_dimensions(
                data_set.data_source_reference
            )
            # Dimension sources may not have any measures -> no aggregation time dimensions.
            if len(aggregation_time_dimensions) == 0:
                source_nodes.append(read_node)
            else:
                for time_dimension_reference in aggregation_time_dimensions:
                    source_nodes.append(
                        MetricTimeDimensionTransformNode[DataSourceDataSet](
                            parent_node=read_node,
                            aggregation_time_dimension_reference=time_dimension_reference,
                        )
                    )

        return source_nodes



"""This module helps to figure out the computational cost for executing a dataflow plan.

There may be multiple possible dataflow plans to realize a set of measures and dimensions (or rather any set of metric
definition instances) because data sets could include an overlapping set of measures and dimensions.

Knowing the cost of a dataflow plan can be used to order the possible plans for optimal execution.
"""


from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Sequence, Generic

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
    SourceDataSetT,
    DataflowPlanNodeVisitor,
    FilterElementsNode,
    WriteToResultDataframeNode,
    OrderByLimitNode,
    ComputeMetricsNode,
    AggregateMeasuresNode,
    JoinAggregatedMeasuresByGroupByColumnsNode,
    JoinOverTimeRangeNode,
    JoinToBaseOutputNode,
    ReadSqlSourceNode,
    WhereConstraintNode,
    CombineMetricsNode,
    ConstrainTimeRangeNode,
    WriteToResultTableNode,
    MetricTimeDimensionTransformNode,
)


class DataflowPlanNodeCost(ABC):
    """Represents the cost to compute the data flow up to a given node."""

    def __lt__(self, other: Any) -> bool:  # type: ignore
        """Implement < so that lists with this can be sorted."""
        if not isinstance(other, DataflowPlanNodeCost):
            return NotImplemented
        return self.as_int < other.as_int

    @property
    @abstractmethod
    def as_int(self) -> int:
        """The cost as an integer for ordering."""
        pass


@dataclass(frozen=True)
class DefaultCost(DataflowPlanNodeCost):
    """Simple cost model where the cost is the number joins * 10 + the number of aggregations"""

    num_joins: int = 0
    num_aggregations: int = 0

    @property
    def as_int(self) -> int:  # noqa: D
        return self.num_joins * 10 + self.num_aggregations

    @staticmethod
    def sum(costs: Sequence[DefaultCost]) -> DefaultCost:  # noqa: D
        return DefaultCost(
            num_joins=sum([x.num_joins for x in costs]),
            num_aggregations=sum([x.num_aggregations for x in costs]),
        )


class DataflowPlanNodeCostFunction(Generic[SourceDataSetT], ABC):
    """A function that calculates the cost for computing the dataflow up to a given node."""

    @abstractmethod
    def calculate_cost(self, node: DataflowPlanNode[SourceDataSetT]) -> DataflowPlanNodeCost:
        """Return the cost for calculating the given dataflow up to the given node."""
        pass


class DefaultCostFunction(
    Generic[SourceDataSetT],
    DataflowPlanNodeCostFunction[SourceDataSetT],
    DataflowPlanNodeVisitor[SourceDataSetT, DefaultCost],
):
    """Cost function using the default cost."""

    def calculate_cost(self, node: DataflowPlanNode[SourceDataSetT]) -> DataflowPlanNodeCost:  # noqa: D
        return node.accept(self)

    def visit_source_node(self, node: ReadSqlSourceNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        # Base case.
        return DefaultCost(num_joins=0, num_aggregations=0)

    def visit_join_to_base_output_node(self, node: JoinToBaseOutputNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        parent_costs = [x.accept(self) for x in node.parent_nodes]

        # Add number of joins to the cost.
        node_cost = DefaultCost(num_joins=len(node.join_targets))
        return DefaultCost.sum(parent_costs + [node_cost])

    def visit_join_aggregated_measures_by_groupby_columns_node(  # noqa: D
        self, node: JoinAggregatedMeasuresByGroupByColumnsNode[SourceDataSetT]
    ) -> DefaultCost:
        parent_costs = [x.accept(self) for x in node.parent_nodes]

        # This node does N-1 joins to link its N parents together
        num_joins = len(node.parent_nodes) - 1
        node_cost = DefaultCost(num_joins=num_joins)
        return DefaultCost.sum(parent_costs + [node_cost])

    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        parent_costs = [x.accept(self) for x in node.parent_nodes]

        # Add the number of aggregations to the cost
        node_cost = DefaultCost(num_aggregations=1)
        return DefaultCost.sum(parent_costs + [node_cost])

    def visit_compute_metrics_node(self, node: ComputeMetricsNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_order_by_limit_node(self, node: OrderByLimitNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_where_constraint_node(self, node: WhereConstraintNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_write_to_result_dataframe_node(  # noqa: D
        self, node: WriteToResultDataframeNode[SourceDataSetT]
    ) -> DefaultCost:
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_write_to_result_table_node(self, node: WriteToResultTableNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_pass_elements_filter_node(self, node: FilterElementsNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_combine_metrics_node(self, node: CombineMetricsNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])

    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode[SourceDataSetT]) -> DefaultCost:  # noqa: D
        parent_costs = [x.accept(self) for x in node.parent_nodes]

        # Add the number of aggregations to the cost (eg 1 per unit time)
        node_cost = DefaultCost(num_aggregations=1)
        return DefaultCost.sum(parent_costs + [node_cost])

    def visit_metric_time_dimension_transform_node(  # noqa: D
        self, node: MetricTimeDimensionTransformNode[SourceDataSetT]
    ) -> DefaultCost:
        return DefaultCost.sum([x.accept(self) for x in node.parent_nodes])



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional, Tuple

from metricflow.time.time_granularity import TimeGranularity

DUNDER = "__"

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class StructuredLinkableSpecName:
    """Parse a qualified name into different parts.

    e.g. listing__ds__week ->
    identifier_links: ["listing"]
    element_name: "ds"
    granularity: TimeGranularity.WEEK
    """

    identifier_link_names: Tuple[str, ...]
    element_name: str
    time_granularity: Optional[TimeGranularity] = None

    @staticmethod
    def from_name(qualified_name: str) -> StructuredLinkableSpecName:
        """Construct from a name e.g. listing__ds__month."""
        name_parts = qualified_name.split(DUNDER)

        # No dunder, e.g. "ds"
        if len(name_parts) == 1:
            return StructuredLinkableSpecName((), name_parts[0])

        associated_granularity = None
        granularity: TimeGranularity
        for granularity in TimeGranularity:
            if name_parts[-1] == granularity.value:
                associated_granularity = granularity

        # Has a time granularity
        if associated_granularity:
            #  e.g. "ds__month"
            if len(name_parts) == 2:
                return StructuredLinkableSpecName((), name_parts[0], associated_granularity)
            # e.g. "messages__ds__month"
            return StructuredLinkableSpecName(tuple(name_parts[:-2]), name_parts[-2], associated_granularity)
        # e.g. "messages__ds"
        else:
            return StructuredLinkableSpecName(tuple(name_parts[:-1]), name_parts[-1])

    @property
    def qualified_name(self) -> str:
        """Return the full name form. e.g. ds or listing__ds__month"""
        items = list(self.identifier_link_names) + [self.element_name]
        if self.time_granularity and self.time_granularity != TimeGranularity.DAY:
            items.append(self.time_granularity.value)
        return DUNDER.join(items)

    @property
    def qualified_name_without_granularity(self) -> str:
        """Return the name without the time granularity. e.g. listing__ds__month -> listing__ds"""
        return DUNDER.join(list(self.identifier_link_names) + [self.element_name])

    @property
    def qualified_name_without_identifier(self) -> str:
        """Return the name without the identifier. e.g. listing__ds__month -> ds__month"""
        return DUNDER.join([self.element_name] + ([self.time_granularity.value] if self.time_granularity else []))

    @property
    def identifier_prefix(self) -> Optional[str]:
        """Return the identifier prefix. e.g. listing__ds__month -> listing"""
        if len(self.identifier_link_names) > 0:
            return DUNDER.join(self.identifier_link_names)

        return None



from __future__ import annotations

from io import StringIO
from typing import Dict, Iterator

import yaml

"""This is the name of the field the SafeLineLoaderWithAddedContext adds to the parsed yaml
   we retrieve line number and the file name from the context stored at this key"""
PARSING_CONTEXT_KEY = "__parsing_context__"


class ParsingContext:
    """Container class for file slice information used to populate model metadata for certain objects"""

    def __init__(self, start_line: int, end_line: int, filename: str, content_node: yaml.Node) -> None:
        """Initializer for the ParsingContext class.

        The contents are represented internally as a yaml.Node in order to allow for lazy serialization to
        string representations.
        """
        self.start_line = start_line
        self.end_line = end_line
        self.filename = filename
        self._content_node = content_node

    @property
    def content(self) -> str:
        """Serialized contents associated with the file slice represented by this ParsingContext object

        This should only be called when a string representation of the contents are needed.
        """
        return yaml.serialize(node=self._content_node)

    def __str__(self) -> str:  # noqa: D
        return f"line: {self.start_line}, filename: {self.filename}"


class YamlConfigLoader:
    """Helper class for loading YAML config strings into an iterator of YAML output"""

    @staticmethod
    def load_all_with_context(name: str, contents: str) -> Iterator:
        """Wraps the yaml.load_all method and returns the resulting iterator with parsing context added to output

        This replaces any calls to yaml.load_all(loader=SafeLineLoaderWithAddedContext), which internally adds
        ParsingContext info. Note PyYAML reads the name property from the input stream IF that input stream is a file
        object, otherwise it replaces it with a constant. Therefore, we use as StringIO instance to pass the contents
        into PyYAML and set the value of the name property on the file object to the name parameter here.
        """
        with StringIO(initial_value=contents) as stream:
            stream.name = name
            for document in yaml.load_all(stream=stream, Loader=SafeLineLoaderWithAddedContext):
                yield document

    @staticmethod
    def is_valid_yaml_file_ending(filename: str) -> bool:
        """Checks if YAML file name ends with one of the supported suffixes"""
        return filename.endswith(".yaml") or filename.endswith(".yml")


class SafeLineLoaderWithAddedContext(yaml.SafeLoader):
    """Adds special field __parsing_context__ to all mappings.

    Credit: https://stackoverflow.com/questions/13319067/parsing-yaml-return-with-line-number
    """

    # we may also want to consider this parser https://yaml.readthedocs.io/en/latest/
    # which supports yaml1.2 and maintains round-trip parsing, for now we use
    # the more established road
    # https://stackoverflow.com/questions/55441300/how-can-i-get-the-parent-node-within-yaml-loader-add-contructor
    def construct_mapping(self, node: yaml.MappingNode, deep: bool = False) -> Dict:
        """Override of the construct_mapping method in the PyYAML SafeLoader class

        This override exists in order to populate the parsing context object with file location
        and raw YAML content information, which will be used to populate the Metadata model construct
        for nodes which request it. The file identifier (node.start_mark.name) is derived from the
        name property on the file-like object passed in as the stream parameter of the load_all call.
        The line numbers are part of the start and end mark properties.

        The content_node is a PyYAML node. We do not serialize it inside the loader because it's a full
        serialization pass on whatever contents this particular mapping node might contain, which could
        be the entire model collection. Rather, we store the node and serialize it on-demand later.

        Note: PyYAML uses metaclasses quite heavily so construct_mapping is in fact defined in the
        SafeConstructor class.
        """
        mapping = super(SafeLineLoaderWithAddedContext, self).construct_mapping(node, deep=deep)
        mapping[PARSING_CONTEXT_KEY] = ParsingContext(
            node.start_mark.line + 1, node.end_mark.line, node.start_mark.name, node
        )  # change to 1-indexed

        return mapping



from __future__ import annotations

import click
import functools
import traceback
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
from pydantic import BaseModel, Extra

from metricflow.instances import (
    DataSourceElementReference,
    DataSourceReference,
    MaterializationModelReference,
    MetricModelReference,
)
from metricflow.model.objects.base import FrozenBaseModel
from metricflow.model.objects.common import Metadata
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.object_utils import assert_values_exhausted

VALIDATE_SAFELY_ERROR_STR_TMPLT = ". Issue occurred in method `{method_name}` called with {arguments_str}"
ValidationContextJSON = Dict[str, Union[str, int, None]]
ValidationIssueJSON = Dict[str, Union[str, int, ValidationContextJSON]]


class ValidationIssueLevel(Enum):
    """Categorize the issues found while validating a MQL model."""

    # Issue should be fixed, but model will still work in MQL
    WARNING = 0
    # Issue doesn't prevent model from working in MQL yet, but will eventually be an error
    FUTURE_ERROR = 1
    # Issue will prevent the model from working in MQL
    ERROR = 2

    @property
    def name_plural(self) -> str:
        """Controlled pluralization of ValidationIssueLevel name value"""

        return f"{self.name}S"


ISSUE_COLOR_MAP = {
    ValidationIssueLevel.WARNING: "cyan",
    ValidationIssueLevel.ERROR: "bright_red",
    ValidationIssueLevel.FUTURE_ERROR: "bright_yellow",
}


class DataSourceElementType(Enum):
    """Maps data source element types to a readable string."""

    MEASURE = "measure"
    DIMENSION = "dimension"
    IDENTIFIER = "identifier"


class FileContext(BaseModel):
    """The base context class for validation issues"""

    file_name: Optional[str]
    line_number: Optional[int]

    class Config:
        """Pydantic class configuration options"""

        extra = Extra.forbid

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""

        context_string = ""

        if self.file_name:
            context_string += f"in file `{self.file_name}`"
            if self.line_number:
                context_string += f" on line #{self.line_number}"

        return context_string

    @classmethod
    def from_metadata(cls, metadata: Optional[Metadata] = None) -> FileContext:
        """Creates a FileContext instance from a Metadata object"""

        return cls(
            file_name=metadata.file_slice.filename if metadata else None,
            line_number=metadata.file_slice.start_line_number if metadata else None,
        )


class MaterializationContext(BaseModel):
    """The context class for validation issues involving materializations"""

    file_context: FileContext
    materialization: MaterializationModelReference

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with materialization `{self.materialization.materialization_name}` {self.file_context.context_str()}"


class MetricContext(BaseModel):
    """The context class for validation issues involving metrics"""

    file_context: FileContext
    metric: MetricModelReference

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with metric `{self.metric.metric_name}` {self.file_context.context_str()}"


class DataSourceContext(BaseModel):
    """The context class for validation issues involving data sources"""

    file_context: FileContext
    data_source: DataSourceReference

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with data source `{self.data_source.data_source_name}` {self.file_context.context_str()}"


class DataSourceElementContext(BaseModel):
    """The context class for validation issues involving dimensions"""

    file_context: FileContext
    data_source_element: DataSourceElementReference
    element_type: DataSourceElementType

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with {self.element_type.value} `{self.data_source_element.element_name}` in data source `{self.data_source_element.data_source_name}` {self.file_context.context_str()}"


ValidationContext = Union[
    FileContext,
    MaterializationContext,
    MetricContext,
    DataSourceContext,
    DataSourceElementContext,
]


class ValidationIssue(ABC, BaseModel):
    """The abstract base ValidationIsssue class that the specific ValidationIssue classes are built from"""

    message: str
    context: Optional[ValidationContext] = None
    extra_detail: Optional[str]

    @property
    @abstractmethod
    def level(self) -> ValidationIssueLevel:
        """The level of of ValidationIssue"""

        raise NotImplementedError

    def as_readable_str(self, verbose: bool = False, prefix: Optional[str] = None) -> str:
        """Return a easily readable string that can be used to log the issue."""
        prefix = prefix or self.level.name

        # The following is two lines instead of one line because
        # technically self.context.context_str() can return an empty str
        context_str = self.context.context_str() if self.context else ""
        context_str += " - " if context_str != "" else ""

        issue_str = f"{prefix}: {context_str}{self.message}"
        if verbose and self.extra_detail is not None:
            issue_str += f"\n{self.extra_detail}"

        return issue_str

    def as_cli_formatted_str(self, verbose: bool = False) -> str:
        """Returns a color-coded readable string for rendering issues in the CLI"""
        return self.as_readable_str(
            verbose=verbose, prefix=click.style(self.level.name, bold=True, fg=ISSUE_COLOR_MAP[self.level])
        )


class ValidationWarning(ValidationIssue, BaseModel):
    """A warning that was found while validating the model."""

    @property
    def level(self) -> ValidationIssueLevel:  # noqa: D
        return ValidationIssueLevel.WARNING


class ValidationFutureError(ValidationIssue, BaseModel):
    """A future error that was found while validating the model."""

    error_date: date

    @property
    def level(self) -> ValidationIssueLevel:  # noqa: D
        return ValidationIssueLevel.FUTURE_ERROR

    def as_readable_str(self, verbose: bool = False, prefix: Optional[str] = None) -> str:
        """Return a easily readable string that can be used to log the issue."""
        return (
            f"{super().as_readable_str(verbose=verbose, prefix=prefix)}"
            f"IMPORTANT: this error will break your model starting {self.error_date.strftime('%b %d, %Y')}. "
        )


class ValidationError(ValidationIssue, BaseModel):
    """An error that was found while validating the model."""

    @property
    def level(self) -> ValidationIssueLevel:  # noqa: D
        return ValidationIssueLevel.ERROR


ValidationIssueType = Union[ValidationWarning, ValidationFutureError, ValidationError]


class ModelValidationResults(FrozenBaseModel):
    """Class for organizating the results of running validations"""

    warnings: Tuple[ValidationWarning, ...] = tuple()
    future_errors: Tuple[ValidationFutureError, ...] = tuple()
    errors: Tuple[ValidationError, ...] = tuple()

    @property
    def has_blocking_issues(self) -> bool:
        """Does the ModelValidationResults have ERROR issues"""
        return len(self.errors) != 0

    @classmethod
    def from_issues_sequence(cls, issues: Sequence[ValidationIssueType]) -> ModelValidationResults:
        """Constructs a ModelValidationResults class from a list of ValidationIssues"""

        warnings: List[ValidationWarning] = []
        future_errors: List[ValidationFutureError] = []
        errors: List[ValidationError] = []

        for issue in issues:
            if issue.level == ValidationIssueLevel.WARNING:
                warnings.append(issue)
            elif issue.level == ValidationIssueLevel.FUTURE_ERROR:
                future_errors.append(issue)
            elif issue.level == ValidationIssueLevel.ERROR:
                errors.append(issue)
            else:
                assert_values_exhausted(issue.level)
        return cls(warnings=tuple(warnings), future_errors=tuple(future_errors), errors=tuple(errors))

    @classmethod
    def merge(cls, results: Sequence[ModelValidationResults]) -> ModelValidationResults:
        """Creates a new ModelValidatorResults instance from multiple instances

        This is useful when there are multiple validators that are run and the
        combined results are desireable. For instance there is a ModelValidator
        and a DataWarehouseModelValidator. These both return validation issues.
        If it's desireable to combine the results, the following makes it easy.
        """

        if not isinstance(results, List):
            results = list(results)

        # this nested comprehension syntax is a little disorienting
        # basically [element for object in list_of_objects for element in object.list_property]
        # translates to "for each element in an object's list for each object in a list of objects"
        warnings = tuple(issue for result in results for issue in result.warnings)
        future_errors = tuple(issue for result in results for issue in result.future_errors)
        errors = tuple(issue for result in results for issue in result.errors)

        return cls(
            warnings=warnings,
            future_errors=future_errors,
            errors=errors,
        )

    @property
    def all_issues(self) -> Tuple[ValidationIssueType, ...]:
        """For when a singular list of issues is needed"""
        return self.errors + self.future_errors + self.warnings

    def summary(self) -> str:
        """Returns a stylized summary string for issues"""

        errors = click.style(
            text=f"{ValidationIssueLevel.ERROR.name_plural}: {len(self.errors)}",
            fg=ISSUE_COLOR_MAP[ValidationIssueLevel.ERROR],
        )
        future_erros = click.style(
            text=f"{ValidationIssueLevel.FUTURE_ERROR.name_plural}: {len(self.future_errors)}",
            fg=ISSUE_COLOR_MAP[ValidationIssueLevel.FUTURE_ERROR],
        )
        warnings = click.style(
            text=f"{ValidationIssueLevel.WARNING.name_plural}: {len(self.warnings)}",
            fg=ISSUE_COLOR_MAP[ValidationIssueLevel.WARNING],
        )
        return f"{errors}, {future_erros}, {warnings}"


def generate_exception_issue(
    what_was_being_done: str,
    e: Exception,
    context: Optional[ValidationContext] = None,
) -> ValidationIssue:
    """Generates a validation issue for exceptions"""
    return ValidationError(
        context=context,
        message=f"An error occured while {what_was_being_done} - {''.join(traceback.format_exception_only(etype=type(e), value=e))}",
        extra_detail="".join(traceback.format_tb(e.__traceback__)),
    )


def _func_args_to_string(*args: Any, **kwargs: Any) -> str:  # type: ignore
    return f"positional args: {args}, key word args: {kwargs}"


def validate_safely(whats_being_done: str) -> Callable:
    """Decorator to safely run validation checks"""

    def decorator_check_element_safely(func: Callable) -> Callable:  # noqa
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> List[ValidationIssueType]:  # type: ignore
            """Safely run a check on model elements"""
            issues: List[ValidationIssueType]
            try:
                issues = func(*args, **kwargs)
            except Exception as e:
                arguments_str = _func_args_to_string(*args, **kwargs)
                issues = [
                    generate_exception_issue(
                        what_was_being_done=whats_being_done
                        + VALIDATE_SAFELY_ERROR_STR_TMPLT.format(
                            method_name=func.__name__, arguments_str=arguments_str
                        ),
                        e=e,
                    )
                ]
            return issues

        return wrapper

    return decorator_check_element_safely


@dataclass(frozen=True)
class DimensionInvariants:
    """Helper object to ensure consistent dimension attributes across data sources.

    All dimensions with a given name in all data sources should have attributes matching these values.
    """

    type: DimensionType
    is_partition: bool


class ModelValidationRule(ABC):
    """Encapsulates logic for checking the values of objects in a model."""

    @staticmethod
    @abstractmethod
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:
        """Check the given model and return a list of validation issues"""
        pass


class ModelValidationException(Exception):
    """Exception raised when validation of a model fails."""

    def __init__(self, issues: Tuple[ValidationIssueType, ...]) -> None:  # noqa: D
        issues_str = "\n".join([x.as_readable_str(verbose=True) for x in issues])
        super().__init__(f"Error validating model. Issues:\n{issues_str}")



from concurrent.futures import ProcessPoolExecutor, as_completed
import copy
import logging
from typing import List, Sequence

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.dir_to_model import ModelBuildResult
from metricflow.model.validations.agg_time_dimension import AggregationTimeDimensionRule
from metricflow.model.validations.data_sources import (
    DataSourceMeasuresUniqueRule,
    DataSourceTimeDimensionWarningsRule,
)
from metricflow.model.validations.dimension_const import DimensionConsistencyRule
from metricflow.model.validations.element_const import ElementConsistencyRule
from metricflow.model.validations.identifiers import (
    IdentifierConfigRule,
    OnePrimaryIdentifierPerDataSourceRule,
    IdentifierConsistencyRule,
)
from metricflow.model.validations.materializations import ValidMaterializationRule
from metricflow.model.validations.metrics import MetricMeasuresRule, CumulativeMetricRule
from metricflow.model.validations.non_empty import NonEmptyRule
from metricflow.model.validations.reserved_keywords import ReservedKeywordsRule
from metricflow.model.validations.unique_valid_name import UniqueAndValidNameRule
from metricflow.model.validations.validator_helpers import (
    ModelValidationResults,
    ValidationIssueType,
    ModelValidationRule,
    ModelValidationException,
)

logger = logging.getLogger(__name__)


class ModelValidator:
    """A Validator that acts on UserConfiguredModel"""

    DEFAULT_RULES = (
        DataSourceMeasuresUniqueRule(),
        DataSourceTimeDimensionWarningsRule(),
        DimensionConsistencyRule(),
        ElementConsistencyRule(),
        IdentifierConfigRule(),
        IdentifierConsistencyRule(),
        OnePrimaryIdentifierPerDataSourceRule(),
        MetricMeasuresRule(),
        CumulativeMetricRule(),
        NonEmptyRule(),
        UniqueAndValidNameRule(),
        ValidMaterializationRule(),
        AggregationTimeDimensionRule(),
        ReservedKeywordsRule(),
    )

    def __init__(self, rules: Sequence[ModelValidationRule] = DEFAULT_RULES, max_workers: int = 1) -> None:
        """Constructor.

        Args:
            rules: List of validation rules to run. Defaults to DEFAULT_RULES
            max_workers: sets the max number of rules to run against the model concurrently
        """

        # Raises an error if 'rules' is an empty sequence or None
        if not rules:
            raise ValueError("ModelValidator 'rules' must be a sequence with at least one ModelValidationRule.")

        self._rules = rules
        self._executor = ProcessPoolExecutor(max_workers=max_workers)

    def validate_model(self, model: UserConfiguredModel) -> ModelBuildResult:
        """Validate a model according to configured rules."""
        model_copy = copy.deepcopy(model)

        issues: List[ValidationIssueType] = []

        futures = [self._executor.submit(validation_rule.validate_model, model_copy) for validation_rule in self._rules]
        for future in as_completed(futures):
            issues.extend(future.result())

        return ModelBuildResult(model=model_copy, issues=ModelValidationResults.from_issues_sequence(issues))

    def checked_validations(self, model: UserConfiguredModel) -> UserConfiguredModel:  # chTODO: remember checked_build
        """Similar to validate(), but throws an exception if validation fails."""
        model_copy = copy.deepcopy(model)
        build_result = self.validate_model(model_copy)

        if build_result.issues.has_blocking_issues:
            raise ModelValidationException(issues=tuple(build_result.issues.all_issues))

        return model



import pytest

from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.test.model.validations.test_unique_valid_name import copied_model
from metricflow.test.test_utils import find_data_source_with


def test_invalid_aggregation_time_dimension(simple_user_configured_model: UserConfiguredModel) -> None:  # noqa:D
    model = copied_model(simple_user_configured_model)
    data_source_with_measures, _ = find_data_source_with(
        model,
        lambda data_source: len(data_source.measures) > 0,
    )

    data_source_with_measures.measures[0].agg_time_dimension = "invalid_time_dimension"

    with pytest.raises(
        ModelValidationException,
        match=(
            "has the aggregation time dimension is set to 'invalid_time_dimension', which not a valid time dimension "
            "in the data source"
        ),
    ):
        model_validator = ModelValidator()
        model_validator.checked_validations(model)
